/* $Id: Generator.java,v 1.1.1.1 2003/02/11 16:19:40 bures Exp $ */
package SOFA.SOFAnode.Made.CodeGen;

import SOFA.SOFAnode.Made.TIR.*;
import java.rmi.RemoteException;
import java.rmi.Naming;
import java.io.IOException;

/** Code generator.
  * 
  * @author Petr Hnetynka
  */
public class Generator {
   Repository defRepository;
   CDLRepository cdlRepository;
   ProfileDef profile;

   ArchitectureDef[] archToGen;   // architectures to generate

   ObjectList objToGen;
   
  /** Creates connection to repository.
    *
    * @exception TIRAccessException  error in access to the TIR
    * @exception RemoteException     remote exception
    * @exception BadIDException      an absolute name in <tt>argv</tt> is in bad format
    * @param  argv  absolute names of architectures to generate
    */
  public Generator(String[] argv) throws RemoteException, TIRAccessException, BadIDException {
    String rmiport = System.getProperty("sofa.rmiport","1099");
    String rmihost = System.getProperty("sofa.rmihost","localhost");
    String profileName = System.getProperty("tir.profile",null);
    try {
      defRepository = (Repository) Naming.lookup("//"+rmihost+":"+rmiport+"/Repository");
    } catch (java.net.MalformedURLException e) {
      throw new TIRAccessException(Res.res.getString("MSG_CantConnTIR")+" --> "+e.getMessage());
    } catch (java.rmi.NotBoundException e) {
      throw new TIRAccessException(Res.res.getString("MSG_CantConnTIR")+" --> "+e.getMessage());
    }
    try {
      cdlRepository = (CDLRepository) defRepository.lookup("cdl","");
      if (profileName != null) {
        profile = defRepository.getProfiles().lookup(profileName);
        if (profile == null)
          throw new TIRAccessException(Res.res.getString("MSG_NoProfile"));
      }
    } catch (TIRExceptLock e) {
      throw new TIRAccessException(Res.res.getString("MSG_LockedTIR")+" --> "+e.getMessage());
    }
    
    if ((System.getProperty("codegen.typesonly","no").compareTo("yes"))==0) {
      // arguments are types, only types will be generated (no holder, helper, ...)
      // codegen.typesonly   ---   undocumented property :-)
      ID[] objIds = stringsToIDs(argv);
      objToGen = getObjectsByID(objIds);
      archToGen = null;
    } else {  // normal behavior - arguments are architectures
      ID[] archIds = stringsToIDs(argv);
      Debug.println(archIds);
      archToGen = getArchitectures(archIds);
      Debug.println(archToGen);
      objToGen = null;
    }
  }

  /** Uses already created connection to TIR.
    *
    * @param rep  reference to TIR
    * @exception TIRAccessException  error in access to the TIR
    * @exception RemoteException     remote exception
    */
  public Generator(Repository rep, ArchitectureDef[] archs) throws RemoteException, TIRAccessException {
    if (rep == null)
      throw new NullPointerException();
    if (archs == null)
      throw new NullPointerException();
    defRepository = rep;
    archToGen = archs;
    try {
      cdlRepository = (CDLRepository) defRepository.lookup("cdl","");
    } catch (TIRExceptLock e) {
      throw new TIRAccessException(Res.res.getString("MSG_LockedTIR")+" --> "+e.getMessage());
    }
  }

  /** Convert string absolute names to ID objects.
    *
    * @param      ids  strings with absolute names
    * @exception  BadIDException  a string with id in bad format
    */
  private ID[] stringsToIDs(String[] ids) throws BadIDException {
    ID[] ret = new ID [ids.length];
    for (int i=0;i<ids.length; i++) {
      ret[i] = new ID(ids[i]);
    }
    return ret;
  }

  /** Return reference to specified object.
    *
    * @param id       id of the object
    * @exception RemoteException  remote exception
    * @exception TIRAccessException  error in access to the TIR
    * @return   reference to object
    */
  private Contained getObjectByName(ID id) throws RemoteException, TIRAccessException {
    try {
      Container con = cdlRepository;
      int i;
      Contained[] a;
      if (id.version()!=null && id.version().compareTo("")!=0) {  // version is set
        for(i=0;i<id.size()-1;i++) {
          a = con.lookup_name(id.get(i));
          if (a==null || a.length==0)
            throw new TIRAccessException(id.toString()+" "+Res.res.getString("MSG_DoesntExist"));
          if (a.length==1) {
            con = (Container) a[0];
          } else {
            boolean found = false;
            for(int j=0;j<a.length;j++) {
              if (a[j].get_identification().version().compareTo(id.version())==0) {
                found = true;
                con = (Container) a[j];
                break;
              }
            }
            if (!found) throw new TIRAccessException(id.toString()+" "+Res.res.getString("MSG_DoesntExist"));
          }
        }
        Contained ret = con.lookup(id.get(i), id.version());
        if (ret==null) 
          throw new TIRAccessException(id.toString()+" "+Res.res.getString("MSG_DoesntExist"));
        return ret;
      } else {  // version is not set
        for(i=0;i<id.size()-1;i++) {
          a = con.lookup_name(id.get(i));
          if (a==null || a.length==0)
            throw new TIRAccessException(id.toString()+" "+Res.res.getString("MSG_DoesntExist"));
          if (a.length==1) {
            con = (Container) a[0];
          } else {  // more objects with same name
            throw new TIRAccessException(Res.res.getString("MSG_MoreVersions")+" "+id.toString()+"\n"+Res.res.getString("MSG_UseProfile"));
          }
        }
        a = con.lookup_name(id.get(i));
        if (a==null || a.length==0)
          throw new TIRAccessException(id.toString()+" "+Res.res.getString("MSG_DoesntExist"));
        if (a.length==1)
          return a[0];
        else  // more versions
          throw new TIRAccessException(Res.res.getString("MSG_MoreVersions")+" "+id.toString()+"\n"+Res.res.getString("MSG_UseProfile"));
      }
      
    } catch (TIRExceptLock e) {
      throw new TIRAccessException(Res.res.getString("MSG_LockedTIR")+" --> "+e.getMessage());
    }
  }

  /** Get architectures from ids. If version is not set, it search in profiles first.
    *
    * @exception RemoteException  remote exception
    * @exception TIRAccessException  error in access to the TIR  
    * @param ids    ids of architectures
    */
  private ArchitectureDef[] getArchitectures(ID[] ids) throws RemoteException, TIRAccessException {
    ArchitectureDef[] ret = new ArchitectureDef [ids.length];
    Contained a;
    try {
      for (int i=0;i<ids.length;i++) {
        if (profile == null) {  // profile was not set
          a = getObjectByName(ids[i]);
          if (a.get_def_kind().value() != DefinitionKind.dk_Architecture)
            throw new TIRAccessException(ids[i].toString()+" "+Res.res.getString("MSG_IsNotArchitecture"));
          ret[i] = (ArchitectureDef) a;
        } else {  // profile is set
          if (ids[i].version()!=null && ids[i].version().compareTo("")!=0) {  // version is set, search directly in objects
            a = getObjectByName(ids[i]);
            if (a.get_def_kind().value() != DefinitionKind.dk_Architecture)
              throw new TIRAccessException(ids[i].toString()+" "+Res.res.getString("MSG_IsNotArchitecture"));
            ret[i] = (ArchitectureDef) a;
          } else {  // search in profile first
            Identification idp = profile.lookup("::cdl"+ids[i].absoluteName());
            if (idp!=null) {  // object is in profile
              Container con = cdlRepository;
              int k = 0;
              for(k=0;k<ids[i].size()-1;k++) {
                Contained[] x = con.lookup_name(ids[i].get(k));
                if (x==null || x.length==0)
                  throw new TIRAccessException(ids[i].toString()+" "+Res.res.getString("MSG_DoesntExist"));
                if (x.length==1) {
                  con = (Container) x[0];
                } else {
                  boolean found = false;
                  for(int j=0;j<x.length;j++) {
                    if (x[j].get_identification().version().compareTo(idp.version())==0) {
                      found = true;
                      con = (Container) x[j];
                      break;
                    }
                  }
                  if (!found) throw new TIRAccessException(ids[i].toString()+" "+Res.res.getString("MSG_DoesntExist"));
                }
              }
              a = con.lookup(ids[i].get(k), idp.version());
              if (a==null)
                throw new TIRAccessException("MSG_ErrorProfile");
              if (a.get_def_kind().value() != DefinitionKind.dk_Architecture)
                throw new TIRAccessException(ids[i].toString()+" "+Res.res.getString("MSG_IsNotArchitecture"));
              ret[i] = (ArchitectureDef) a;
            } else {  // object is not in profile
              a = getObjectByName(ids[i]);
              if (a.get_def_kind().value() != DefinitionKind.dk_Architecture)
                throw new TIRAccessException(ids[i].toString()+" "+Res.res.getString("MSG_IsNotArchitecture"));
              ret[i] = (ArchitectureDef) a;
            }
          }
        }
      }
    } catch (TIRExceptLock e) {
      throw new TIRAccessException(Res.res.getString("MSG_LockedTIR")+" --> "+e.getMessage());
    }
    return ret;
  }

  /** absolute name in string converts to ArrayList */
  private java.util.ArrayList absNameToList(String name) {
    java.util.ArrayList ret = new java.util.ArrayList(10);
    int i;
    int count = 0;
    for (i=0;i<name.length();i++) {
      if (name.charAt(i)==':' && name.charAt(i+1)==':') 
        count++;
    }
    int j = 0;
    for(i=0;i<count;i++) {
      StringBuffer str = new StringBuffer();
      while (name.charAt(j)==':')
        j++;
      while (name.charAt(j)!=':' && j<name.length()) {
        str = str.append(name.charAt(j));
        j++;
      }
      ret.add(str.toString());
    }
    if (ret.size()==0)
      return null;
    return ret;
  }

  /** Find frame of instance. It goes through typedefs and arrays.
    *
    * @param     inst    instance from which frame is searched 
    * @exception RemoteException  remote exception 
    */
  CDLType getFrameOfInst(InstDef inst) throws RemoteException {
    CDLType obj = inst.type();
    int defKind = obj.get_def_kind().value();
    while (defKind == DefinitionKind.dk_Typedef || defKind == DefinitionKind.dk_Array) {
      if (defKind == DefinitionKind.dk_Typedef) 
        obj = ((TypedefDef) obj).original_type();
      else
        obj = ((ArrayDef) obj).element_type();
      defKind = obj.get_def_kind().value();
    }
    return obj;
  }

  /** Returns list with all frames, which must be generated.
    *
    * @param     obj    architecture from which frames are searched 
    * @exception RemoteException  remote exception 
    * @exception TIRAccessException  error in access to the TIR
    */
  private ObjectList getAllFrames(ArchitectureDef obj) throws RemoteException, TIRAccessException {
    ObjectList list = new ObjectList();
    list.add(obj.frame());
    Contained[] cont = null;
    try {
      // get instances only
      cont = obj.contents(defRepository.get_spec_def_kind(DefinitionKind.dk_Inst));
    } catch (TIRExceptLock e) {
      throw new TIRAccessException(Res.res.getString("MSG_LockedTIR")+" --> "+e.getMessage());
    }
    for (int i=0; i<cont.length; i++) {
      CDLType frame = getFrameOfInst((InstDef) cont[i]);
      if (frame.get_def_kind().value() != DefinitionKind.dk_Frame) {
        Identification idl = obj.get_identification();
        throw new TIRAccessException(Res.res.getString("MSG_InstTIRError")+" "+idl.absolute_name().name()+":"+idl.version());
      }
      list.add(frame);
    }
    return list;
  }

  /** Gets list with all objects, which must be generated.
    *
    * @param     frames    list of frames from which objects are searched 
    * @exception RemoteException  remote exception 
    * @exception TIRAccessException  error in access to the TIR
    */
  private ObjectList getAllObjects(ObjectList frames) throws RemoteException, TIRAccessException {
    ObjectList types = new ObjectList();
    try {
      int i;
      for (i=0;i<frames.size();i++) {  // add all objects from frames
        Contained[] cont = ((Container)frames.getObject(i)).contents(null);
        int j;
        for (j=0;j< cont.length; j++) {
          firstLevelOfScaning = true;
          scanObject(cont[j], types);
        }
      }
      for(i=0;i<types.size(); i++) {
        firstLevelOfScaning = true;
        scanObject(types.getObject(i), types);
      }
    } catch (TIRExceptLock e) {
      throw new TIRAccessException(Res.res.getString("MSG_LockedTIR")+" --> "+e.getMessage());
    }
    return types;
  }

  
  private boolean firstLevelOfScaning;

  /** Add all objects that are in <tt>obj</tt> to the <tt>list</tt>.  
    *
    * @param     obj  scanned object
    * @param     list add all types to this list
    * @exception RemoteException  remote exception 
    * @exception TIRAccessException  error in access to the TIR
    */
  private void scanObject(TIRObject obj, ObjectList list) throws RemoteException, TIRAccessException {
    try {
      if (obj instanceof Contained) {
        Debug.println("Enter scanObject with: "+((Contained)obj).get_identification().name());
//        System.out.println(firstLevelOfScaning);
      }
      Debug.println(list);
      
      if (!firstLevelOfScaning  && (obj instanceof Contained) && list.containsObject((Contained)obj)) {
//        System.out.println("*** End (break) of \"scanObject\" method "+((Contained)obj).get_identification().name());
        return;
      }

      firstLevelOfScaning = false;
      
      int i;
      int defKind;
      Contained[] cont;
      Container defIn;
      Contained cobj;
      CDLType tp, newtp;
      boolean scan = true;
      switch (obj.get_def_kind().value()) {
      case DefinitionKind.dk_Attribute:
        cobj = (Contained) obj;
        tp = ((AttributeDef) obj).type();
        if (tp.get_def_kind().value() != DefinitionKind.dk_Primitive) {
          if (tp instanceof Contained) {
            if (!list.containsObject((Contained) tp)) {
              list.addObject((Contained) tp);
            } else {
              scan = false;
            }
          }
          if (scan)
            scanObject( tp, list);
        }
        break;
      case DefinitionKind.dk_Constant:
        cobj = (Contained) obj;
        tp = ((ConstantDef) obj).type();
        if (tp.get_def_kind().value() != DefinitionKind.dk_Primitive) {
          if (tp instanceof Contained) {
            if (!list.containsObject((Contained) tp)) {
              list.addObject((Contained) tp);
            } else {
              scan = false;
            }
          }
          if (scan)
            scanObject( tp, list);
        }
        Expressions.scanExpr(((ConstantDef)obj).value(),list);
        if (! (cobj.get_defined_in() instanceof InterfaceDef))  // add only constants not within an interface
          list.addObject(cobj);
        break;
      case DefinitionKind.dk_Interface:
        cobj = (Contained) obj;
        cont = ((Container)obj).contents(null);
        for (i=0;i<cont.length;i++) {
          scanObject(cont[i], list);
        }
        InterfaceDef[] bIfaces = ((InterfaceDef) obj).base_interfaces();
        for (i=0;i<bIfaces.length;i++) {
          list.addObject(bIfaces[i]);
        }
        list.addObject(cobj);
        break;
      case DefinitionKind.dk_Operation:
        cobj = (Contained) obj;
        tp = ((OperationDef) obj).result();
        if (tp.get_def_kind().value() != DefinitionKind.dk_Primitive) {
          if (tp instanceof Contained) {
            if (!list.containsObject((Contained) tp)) {
              list.addObject((Contained) tp);
            } else {
              scan = false;
            }
          }
          if (scan)
            scanObject( tp, list);
        }
        ParamDescr[] par = ((OperationDef) obj).params();
        for(i=0;i<par.length;i++) {
          tp = par[i].type();
          if (tp.get_def_kind().value() != DefinitionKind.dk_Primitive) {
            if (tp instanceof Contained) {
              if (!list.containsObject((Contained) tp)) {
                list.addObject((Contained) tp);
              } else {
                scan = false;
              }
            }
            if (scan)
              scanObject( tp, list);
          }
        }
        ExceptionDef[] excs = ((OperationDef) obj).exceptions();
        Contained newcont;
        for (i=0;i<excs.length; i++) {
          newcont = excs[i];
          defIn = newcont.get_defined_in();
          while (!(defIn instanceof ModuleDef || defIn instanceof Repository)) {
            newcont = (Contained) defIn;
            defIn = newcont.get_defined_in();
          }
          list.addObject(newcont);
          scanObject(excs[i], list);
        }
        break;
      case DefinitionKind.dk_Typedef:
        cobj = (Contained) obj;
        tp = ((TypedefDef) obj).original_type();
        if (tp.get_def_kind().value() != DefinitionKind.dk_Primitive) {
          if (tp instanceof Contained) {
            if (!list.containsObject((Contained) tp)) {
              list.addObject((Contained) tp);
            } else {
              scan = false;
            }
          }
          if (scan)
            scanObject( tp, list);
        }
        list.addObject(cobj);
        break;
      case DefinitionKind.dk_Exception:
      case DefinitionKind.dk_Struct:
        cobj = (Contained) obj;
        cont = ((Container)obj).contents(null);
        for (i=0;i<cont.length;i++) {
          scanObject(cont[i], list);
        }
        defIn = cobj.get_defined_in();
        if (defIn instanceof ModuleDef || defIn instanceof Repository ) {
          list.addObject(cobj);
        }
        break;
      case DefinitionKind.dk_Union:
        cobj = (Contained) obj;
        cont = ((Container)obj).contents(null);
        for (i=0;i<cont.length;i++) {
          scanObject(cont[i], list);
        }
        tp = ((UnionDef) obj).switch_type();
        if (tp.get_def_kind().value() != DefinitionKind.dk_Primitive) {
          list.addObject((Contained) tp);
          scanObject( tp, list);
        }
        defIn = cobj.get_defined_in();
        if (defIn instanceof ModuleDef || defIn instanceof Repository )
          list.addObject(cobj);
        break;
      case DefinitionKind.dk_Enum:
        cobj = (Contained) obj;
        defIn = cobj.get_defined_in();
        if (defIn instanceof ModuleDef || defIn instanceof Repository )
          list.addObject(cobj);
        break;
      case DefinitionKind.dk_Sequence:
        tp = ((SequenceDef) obj).element_type();
        if (tp.get_def_kind().value() != DefinitionKind.dk_Primitive) {
          if (tp instanceof Contained) {
            if (!list.containsObject((Contained) tp)) {
              list.addObject((Contained) tp);
            } else {
              scan = false;
            }
          }
          if (scan)
            scanObject( tp, list);
        }
        Expressions.scanExpr(((SequenceDef) obj).bound(),list);
        break;
      case DefinitionKind.dk_Array:
        tp = ((ArrayDef) obj).element_type();
        if (tp.get_def_kind().value() != DefinitionKind.dk_Primitive) {
          if (tp instanceof Contained) {
            if (!list.containsObject((Contained) tp)) {
              list.addObject((Contained) tp);
            } else {
              scan = false;
            }
          }
          if (scan)
            scanObject( tp, list);
        }
        Expressions.scanExpr(((ArrayDef) obj).length(),list);
        break;
      case DefinitionKind.dk_Frame: 
        cont = ((Container)obj).contents(null);
        for (i=0;i<cont.length;i++) {
          scanObject(cont[i], list);
        }
        break;
      case DefinitionKind.dk_Provides:
        list.addObject( (Contained) ((ProvideDef) obj).provide());
        break;
      case DefinitionKind.dk_Requires: 
        list.addObject( (Contained) ((RequireDef) obj).require());
        break;
      case DefinitionKind.dk_Property:
        cobj = (Contained) obj;
        tp = ((PropertyDef) obj).type();
        if (tp.get_def_kind().value() != DefinitionKind.dk_Primitive) {
          if (tp instanceof Contained) {
            if (!list.containsObject((Contained) tp)) {
              list.addObject((Contained) tp);
            } else {
              scan = false;
            }
          }
          if (scan)
            scanObject( tp, list);
        }
        // list.addObject(cobj);
        break;
      case DefinitionKind.dk_StructMember:
        tp = ((StructMember) obj).type();
        if (tp.get_def_kind().value() != DefinitionKind.dk_Primitive) {
          if (tp instanceof Contained) {
            newtp = tp;
            defIn = ((Contained) newtp).get_defined_in();
            while (!(defIn instanceof ModuleDef || defIn instanceof Repository)) {
              newtp = (CDLType) defIn;
              defIn = ((Contained) newtp).get_defined_in();
            }
            list.addObject((Contained) newtp);
          }
          scanObject( tp, list);
        }
        break;
      case DefinitionKind.dk_UnionMember:
        tp = ((UnionMember) obj).type();
        if (tp.get_def_kind().value() != DefinitionKind.dk_Primitive) {
          if (tp instanceof Contained) {
            newtp = tp;
            defIn = ((Contained) newtp).get_defined_in();
            while (!(defIn instanceof ModuleDef || defIn instanceof Repository)) {
              newtp = (CDLType) defIn;
              defIn = ((Contained) newtp).get_defined_in();
            }
            list.addObject((Contained) newtp);
          }
          scanObject( tp, list);
        }
        ExprOperDef[] exprs = ((UnionMember) obj).label();
        for (i=0;i<exprs.length;i++) {
          Expressions.scanExpr(exprs[i],list);
        }
        break;
      case DefinitionKind.dk_String:
        Expressions.scanExpr(((StringDef) obj).bound(),list);
        break;
      case DefinitionKind.dk_Wstring:
        Expressions.scanExpr(((WstringDef) obj).bound(),list);
        break;
      }


      
//      if (obj instanceof Contained)
//        System.out.println("*** End of \"scanObject\" method "+((Contained)obj).get_identification().name());
        
    } catch (TIRExceptLock e) {
      throw new TIRAccessException(Res.res.getString("MSG_LockedTIR")+" --> "+e.getMessage());
    }
  }

  /** Run generation of code.
    *
    * @exception RemoteException  remote exception
    * @exception TIRAccessException  error in access to the TIR
    */
  public void generate() throws RemoteException, TIRAccessException, IOException, GenerException {
    JavaGenerator jgen;
    if (archToGen != null) {
      // normal (full) generation
      ObjectList allFrames = new ObjectList();
      ObjectList fr;
      int i;
      for (i=0;i<archToGen.length;i++) {
        fr = getAllFrames(archToGen[i]);
        allFrames.mergeTogether(fr);
      }
      Debug.println(allFrames);

      ObjectList allObjects = getAllObjects(allFrames);
      Debug.println(allObjects);

      jgen = new JavaGenerator(allObjects, archToGen);
    } else {
      // typesonly generation
      Debug.println(objToGen);
      ObjectList allObjects = getAllTypeObjects(objToGen);
      Debug.println(allObjects);
      jgen = new JavaGenerator(allObjects, true);
    }
    jgen.generate();
  }

  /** Get objects from ids. If version is not set, it search in profiles first.
    * This method is used only, when codegen.typesonly property is used.
    *
    * @exception RemoteException  remote exception
    * @exception TIRAccessException  error in access to the TIR  
    * @param ids    ids of object
    */
  private ObjectList getObjectsByID(ID[] ids) throws RemoteException, TIRAccessException {
    ObjectList ret = new ObjectList();
    Contained a;
    try {
      for (int i=0;i<ids.length;i++) {
        if (profile == null) {  // profile was not set
          a = getObjectByName(ids[i]);
          ret.addObject(a);
        } else {  // profile is set
          if (ids[i].version()!=null && ids[i].version().compareTo("")!=0) {  // version is set, search directly in objects
            a = getObjectByName(ids[i]);
            ret.addObject(a);
          } else {  // search in profile first
            Identification idp = profile.lookup("::cdl"+ids[i].absoluteName());
            if (idp!=null) {  // object is in profile
              Container con = cdlRepository;
              int k = 0;
              for(k=0;k<ids[i].size()-1;k++) {
                Contained[] x = con.lookup_name(ids[i].get(k));
                if (x==null || x.length==0)
                  throw new TIRAccessException(ids[i].toString()+" "+Res.res.getString("MSG_DoesntExist"));
                if (x.length==1) {
                  con = (Container) x[0];
                } else {
                  boolean found = false;
                  for(int j=0;j<x.length;j++) {
                    if (x[j].get_identification().version().compareTo(idp.version())==0) {
                      found = true;
                      con = (Container) x[j];
                      break;
                    }
                  }
                  if (!found) throw new TIRAccessException(ids[i].toString()+" "+Res.res.getString("MSG_DoesntExist"));
                }
              }
              a = con.lookup(ids[i].get(k), idp.version());
              if (a==null)
                throw new TIRAccessException("MSG_ErrorProfile");
              ret.addObject(a);
            } else {  // object is not in profile
              a = getObjectByName(ids[i]);
              ret.addObject(a);
            }
          }
        }
      }
    } catch (TIRExceptLock e) {
      throw new TIRAccessException(Res.res.getString("MSG_LockedTIR")+" --> "+e.getMessage());
    }
    return ret; 
  }

  /** Gets list with all objects, which must be generated.
    * Used only, when codegen.typesonly property is used.
    *
    * @param     obj    list of objects from which objects are searched 
    * @exception RemoteException  remote exception 
    * @exception TIRAccessException  error in access to the TIR
    */
  private ObjectList getAllTypeObjects(ObjectList obj) throws RemoteException, TIRAccessException {
    int i;
    for(i=0;i<obj.size(); i++) {
      firstLevelOfScaning = true;
      scanObject(obj.getObject(i), obj);
    }
    return obj;
  }
}
