/**
 * BundleImpl.java is a part of the SOFA project.
 * This file was created by pepan on 19.3.2003.
 */
package SOFA.SOFAnode.TR.Impl;

import SOFA.SOFAnode.InOut.Bundle;
import SOFA.SOFAnode.InOut.InOutException;
import SOFA.SOFAnode.TR.ComponentInfo;
import SOFA.Util.Tools;
import SOFA.Util.VMProperties;
import cz.cuni.sofa.lib.InputStream;
import cz.cuni.sofa.lib.OutputStream;
import cz.cuni.sofa.lib.*;

import java.io.*;
import java.util.*;
import java.util.jar.*;
import java.lang.Object;

/**
 * A bundle wrapping a set of components into a single zipped JAR file (actually, an arbitrary
 * stream is used). Such a bundle is intented to be sent between two SOFA nodes when exchanging
 * components. The manifest file of such a JAR file contains a description of all components
 * stored in the bundle. If a bundle manifest file describes a component, but the bundle does not
 * contain the component files, the bundle is called a meta-bundle.
 * @author Petr Panuska
 */
public class BundleImpl implements Bundle, Serializable, Streamable {

  /**
   * All components wrapped in this bundle.
   */
  private Hashtable components;

  /**
   * The URL pointing to the local SOFA node where this bundle was created. It's used
   * when storing the description of components only, to denote where the component
   * can be obtained from.
   */
  private static String sofaNodeName;

  /**
   * Points to the root directory of files created just temporarily when unpacking the
   * bundle from a stream and just before the installation to TR.
   */
  private File tempDir = null;

  /**
   * This instance is used by the bundle when it denotes that a particular component does not contain
   * any single file - it is a component description only. Such a bundle is called a meta-bundle.
   */
  public static final ComponentFiles NO_FILES = new ComponentFiles();

  /**
   * Creates an empty bundle.
   */
  public BundleImpl () {
    components = new Hashtable();
    sofaNodeName = "sofa://" + System.getProperty(VMProperties.NODE_NAME);
  }

  /**
   * Gets an array of all components currently stored in this bundle.
   * @return an array of component info objects.
   */
  public ComponentInfo[] getComponents () {
    Set keySet = components.keySet();
    int size = keySet.size();
    ComponentInfo[] comps = new ComponentInfo[size];
    keySet.toArray(comps);
    return comps;
  }

  /**
   * Returns <code>true</code> if the asked component description is stored in this bundle.
   * @param descr the specified component.
   * @return whether the component description is contained in this bundle.
   */
  public boolean contains (ComponentInfo descr) {
    return components.containsKey(descr);
  }

  /**
   * Returns <code>true</code> if the asked component itself
   * (not only the description) is contained withing this bundle.
   * @param comp the specified component.
   * @return whether the component itself is contained in this bundle.
   */
  public boolean containsComponent (ComponentInfo comp) {
    ComponentFiles files = (ComponentFiles) components.get(comp);
    return (files != null && files != NO_FILES);
  }

  /**
   * Returns files creating the asked component. The object wrapping all these files is of
   * type {@link ComponentFiles}.
   * @param descr the component being asked.
   * @return files creating the component.
   * @throws InOutException when the asked component is not stored in this bundle.
   */
  public java.lang.Object getComponent (ComponentInfo descr) throws InOutException {
    ComponentFiles files = (ComponentFiles) components.get(descr);
    if (files == null)
      throw new InOutException("Required component has not been stored in this bundle. " + descr);
    return files;
  }

  /**
   * Adds files pertaining a component to the internal data structure of this bundle.
   * @param comp the component.
   * @param files the files pertaining the component.
   */
  void add (ComponentInfoImpl comp, ComponentFiles files) {
    comp.setLocation(comp.getName() + comp.getImplementationVersion());
    components.put(comp, files);
  }

  /**
   * Loads this bundle from a stream. It loads all files previously stored in a zipped
   * JAR file including the manifest file and restores all internal data structures of
   * this bundle. It uses a temporarily created directory for storing component files.
   * @param stream the stream this bundle is being loaded from.
   */
  public void loadFromJar (BundleInputStream stream) throws IOException {
    tempDir = stream.getTempRootDirectory(); // sets the root of temporary files
    Manifest manifest = stream.getManifest();

    Map entries = manifest.getEntries();
    // process normal components (have files stored in the stream)
    for (BundleEntry entry = stream.getNextBundleEntry(); entry != null; entry = stream.getNextBundleEntry()) {
      Attributes attrs = (Attributes) entries.remove(entry.getName());  // removes entries from the map

      ComponentInfoImpl compInfo = getComponentInfo(attrs);
      components.put(compInfo, entry.getFiles());

    }
    stream.close();
    if (!entries.isEmpty()) {   // process meta-components
      Iterator iter = entries.keySet().iterator();
      while (iter.hasNext()) {
        String key = (String) iter.next();
        Attributes attrs = (Attributes) entries.get(key);

//        Attributes attrs = (Attributes) iter.next();

        ComponentInfoImpl compInfo = getComponentInfo(attrs);
        components.put(compInfo, NO_FILES);
      }
    }
  }

  /**
   * Gets a component described by the specified attributes. These attributes are taken
   * from the manifest file.
   * @param attrs the specified attributes.
   * @return the component info with no files assigned.
   */
  private ComponentInfoImpl getComponentInfo (Attributes attrs) {
    String name = attrs.getValue("Component-Name");
    String version = attrs.getValue("Implementation-Version");
    String location = attrs.getValue("Component-Path");

    return new ComponentInfoImpl(name, version, location);
  }

  /**
   * Returns the content of the manifest file of this bundle.
   * @return the manifest file of this bundle.
   */
  private String createManifest () {

    StringBuffer manifest = new StringBuffer();
    manifest.append("Manifest-Version: 1.0\n");
    manifest.append("Created-By: ");
    manifest.append(System.getProperty(VMProperties.NODE_NAME, "Unknown node"));
    manifest.append("\n");

    manifest.append("\n");

    class MyVector extends Vector {  // implements naming directories for stored components with different names (even if the components have the same name)
      String getComponentPath (ComponentInfo info) {
        String name = info.getName();
        String shortName = name.substring(name.lastIndexOf("::") + 2); // takes the short name of a component (without modules)
        String finalName = shortName;
        for (int i = 2; this.contains(finalName); i++) { // tries to find different name
          finalName = shortName + Integer.toString(i);     // just adds a number to the original name
        }
        this.add(finalName);          // rememeber the name for next use
        return finalName;
      }
    }
    ;

    MyVector compDirs = new MyVector();

    Enumeration comp = components.keys();
    while (comp.hasMoreElements()) {
      ComponentInfoImpl info = (ComponentInfoImpl) comp.nextElement();
      ComponentFiles compFiles = (ComponentFiles) components.get(info); // component files
      String name = compDirs.getComponentPath(info);
      String location = (compFiles == NO_FILES) ? sofaNodeName : name;
      info.setLocation(location);
      manifest.append("Name: ");
      manifest.append(name);
      manifest.append("\n");
      manifest.append("Component-Name: ");
      manifest.append(info.getName());
      manifest.append("\n");
      manifest.append("Implementation-Version: ");
      manifest.append(info.getImplementationVersion());
      manifest.append("\n");
      manifest.append("Component-Path: ");
      manifest.append(location);
      manifest.append("\n");
      manifest.append("\n");
      manifest.append("\n");
    }

    return manifest.toString();

  }

  /**
   * Saves this bundle to the specified stream. It stores all component files to
   * a zipped directory with a manifest file - a JAR file. The manifest file describes
   * all components being stored.
   * @param stream the stream where is this bundle being stored to.
   * @throws IOException when some IO operation fails.
   */
  public void saveToJar (BundleOutputStream stream) throws IOException {
    stream.putNextEntry(new BundleEntry(JarFile.MANIFEST_NAME));
    String manifest = createManifest();
    byte[] buffer = manifest.getBytes();
    stream.write(buffer);

    Enumeration comp = components.keys();
    while (comp.hasMoreElements()) {
      ComponentInfoImpl info = (ComponentInfoImpl) comp.nextElement();  // the component
      ComponentFiles compFiles = (ComponentFiles) components.get(info); // component files
      if (compFiles != NO_FILES) // don't store files belonging to a meta-component
        stream.putNextBundleEntry(new BundleEntry(info.getLocation(), compFiles)); // location here serves as a prefix of all component file names
    }

    stream.close();
  }

  /**
   * Reads a bundle from a stream.
   * @param is the input stream this bundle is being read from.
   * @throws IOException when some IO operation fails.
   * @see #loadFromJar
   */
  public void _read (InputStream is) throws IOException {
/*
    //todo following code is here for testing purposes only
    FileOutputStream f = new FileOutputStream("c:\\ww\\tmp2.zip");
    final int BUFFER_SIZE = 4096;
    byte[] buffer = new byte[BUFFER_SIZE];
    for (int length;                                     // use buffer to copy the entry to the file
         (length = is.read(buffer, 0, BUFFER_SIZE)) >= 0;
         f.write(buffer, 0, length));
    f.close();
    FileInputStream is2 = new FileInputStream("c:\\ww\\tmp2.zip");
    loadFromJar(new BundleInputStream(is2));
*/
    loadFromJar(new BundleInputStream(is));
  }

  /**
   * Writes this bundle to a stream.
   * @param os the output stream this bundle is being stored to.
   * @throws IOException when some IO operation fails.
   * @see #saveToJar
   */
  public void _write (OutputStream os) throws IOException {
/*    //todo following code is here for testing purposes only
    FileOutputStream f = new FileOutputStream("c:\\ww\\tmp3.zip");
    saveToJar(new BundleOutputStream(f));
    f.close();
*/
    saveToJar(new BundleOutputStream(os));
  }

  /**
   * Deletes all temporarily created files and directories. A bundle is a structure created
   * temporarily only. If it's going to be garbage collected, it deletes the temporary
   * files as well.
   */
  protected void finalize () throws Throwable {
    if (tempDir != null) // if there were some temporary files, delete these files
      Tools.deleteSubTree(tempDir);
    super.finalize();
  }

}
