/**
 * TRImpl.java is a part of the SOFA project.
 * This file was created by pepan on 19.3.2003.
 */
package SOFA.SOFAnode.TR.Impl;

import SOFA.SOFAnode.InOut.Bundle;
import SOFA.SOFAnode.Run.Deployment.DeploymentDescriptorImpl;
import SOFA.SOFAnode.TR.*;
import SOFA.SOFAnode.TR.Connector.CNTR2InOut;
import SOFA.Util.Tools;

import javax.xml.parsers.ParserConfigurationException;
import java.io.*;
import java.rmi.RemoteException;
import java.util.*;

import org.xml.sax.SAXException;

/**
 * The implementation of the Template Repository (TR). TR is used for storing components
 * that are installed on a local SOFA node. It employs the standard file system.
 * @author Petr Panuska
 */
public class TRImpl implements CNTR2InOut, TR2InOut {

  /**
   * The root directory for storing content of this Template Repository.
   */
  private String dir;

  /**
   * Data structure for storing all installed components.
   */
  private class ComponentStore {

    /**
     * The collection of stored components. Each key as well as the associated value is an instance of
     * {@link ComponentInfo}. It's used for getting the proper instance of <code>ComponentInfo</code>.
     */
    private Hashtable componentInfos;

    /**
     * The collection of component files.
     * Each key is an instance of {@link ComponentInfo} class, the assigned value is an instance
     * of {@link ComponentFiles} class - a list of files the components consist of.
     */
    private Hashtable componentFiles;

    /**
     * Creates instances of {@link #componentInfos} and {@link #componentFiles}.
     */
    private ComponentStore () {
      componentInfos = new Hashtable();
      componentFiles = new Hashtable();
    }

    /**
     * Adds one component to this store and assigns the associated component files.
     * @param info the component.
     * @param files the list of files the component consists of.
     */
    private void addFiles (ComponentInfo info, ComponentFiles files) {
      componentInfos.put(info, info);
      componentFiles.put(info, files);
    }

    /**
     * Gets component files the given component consists of.
     * @param info the given component.
     * @return files the given component consists of.
     */
    private ComponentFiles getFiles (ComponentInfo info) {
      return (ComponentFiles) componentFiles.get(info);
    }

    /**
     * Gets the instance of {@link ComponentInfoImpl} stored in this component store.
     * @param info component info object where the component name and version are filled only.
     * @return the instance stored in this store.
     */
    private ComponentInfoImpl getInfo (ComponentInfo info) {
      return (ComponentInfoImpl) componentInfos.get(info);
    }

    /**
     * Gets the instance of {@link ComponentInfoImpl} stored in this component store.
     * @param fullName full name of the component (in format <em>name[version]</em>).
     * @return the instance stored in this store.
     */
    private ComponentInfoImpl getInfo (String fullName) {
      return getInfo(ComponentInfoImpl.fromString(fullName));
    }

    /**
     * Gets all components stored in this store.
     * @return set of components.
     */
    private Set getInfos () {
      return componentInfos.keySet();
    }

    /**
     * Is the given component stored in this component store?
     * @param info the given component.
     * @return <code>true</code> if the component is stored in here.
     */
    private boolean contains (ComponentInfo info) {
      return componentInfos.contains(info);
    }

    /**
     * Sets {@link ComponentInfoImpl#subComponents} of the given component. It allows for
     * the existence of all sub-components in this store when performing this operation.
     * If some sub-component is not stored in here, it throws the {@link TRException}.
     * @param info the component its sub-components are going to be set.
     * @throws IOException if some I/O operation fails.
     * @throws ParserConfigurationException if the component assembly file is corrupted or missing.
     * @throws SAXException if the component assembly file is corrupted or missing.
     */
    void setSubComponents (ComponentInfoImpl info) throws IOException, ParserConfigurationException, SAXException {
      ComponentFiles files = (ComponentFiles) componentFiles.get(info);
      if (files == null)  // no files associated
        return;
      AssemblyDescriptorImpl ad = new AssemblyDescriptorImpl(files.getAssemblyDescriptorFile());
      Set set = ad.getSubComponents();  // the set of strings decribing subcomponents of the component info
      Set subComponents = new HashSet();// set of subcomponets (set of ComponentInfoImpl objects)
      for (Iterator iterator = set.iterator(); iterator.hasNext();) {
        String fullName = (String) iterator.next();
        iterator.remove();
        ComponentInfo comp = getInfo(fullName);
        subComponents.add(comp);
      }

      info.setSubComponents((ComponentInfoImpl[]) subComponents.toArray(new ComponentInfoImpl[subComponents.size()]));

      if (!set.isEmpty()) // if the set is not empty, it means that some subcomponents are missing in the TR
        throw new TRException("Some of components being processed use sub-components which are not stored in this TR.");

    }
  }

  /**
   * The set of all components installed in this TR.
   */
  private ComponentStore components;

  /**
   * A list of installed components.
   */
  private Properties implIndex;

  /**
   * A list of installed interfaces and types.
   */
  private Properties ifaceIndex;

  /**
   * A path to the index with installed components.
   */
  private String implPath;

  /**
   * A path to the index with installed interfaces.
   */
  private String ifacePath;

  /**
   * Creates the template repository.
   * @param dir the directory where contents of this repository is stored.
   * Since TR needs a write access to this directory, it must be a local directory.
   */
  public TRImpl (String dir) throws RemoteException, TRException {
    this.dir = dir;
    init();
  }

  /**
   * Initializes internal structures of TR.
   * @throws TRException when some error occurs (some necessary file is missing, etc.)
   */
  private synchronized void init () throws TRException {
//    componentFiles = new Hashtable();
//    components = new Hashtable();
    components = new ComponentStore();

    //create indexes
    implIndex = new Properties();
    ifaceIndex = new Properties();
    implPath = dir + File.separator + "impl" + File.separator;
    ifacePath = dir + File.separator + "iface" + File.separator;
    try {
      FileInputStream fis = new FileInputStream(implPath + "info");
      implIndex.load(fis);
      fis.close();

      fis = new FileInputStream(ifacePath + "index");
      ifaceIndex.load(fis);
      fis.close();

    } catch (java.io.IOException e) {
      System.out.println("IOException: " + e.getMessage());
      System.exit(1);
    }

    // look for all components stored in TR
//    Enumeration enum = implIndex.keys();
    Set collection = implIndex.keySet();
//    while (enum.hasMoreElements()) {
    for (Iterator iterator = collection.iterator(); iterator.hasNext();) {
//      String nameVer = (String) enum.nextElement(); // component name including implementation version
      String nameVer = (String) iterator.next(); // component name including implementation version

      String name = nameVer.substring(0, nameVer.lastIndexOf('[')); // component name
      String path = fullNameToFileName(nameVer); // relative path to the component files


      AssemblyDescriptorImpl ad = null;
      try {
        ad = new AssemblyDescriptorImpl(new File(dir + File.separator + "impl" + path + "index.dc"));
      } catch (Exception e) {
        throw new TRException("Exception thrown when initializing TR!\n" +
          "When creating the assembly descriptor!\n" + path + "\n" + e, e);
      }
      ComponentInfo info = new ComponentInfoImpl(name, ad.getImplementationVersion());
      ComponentFiles files = new ComponentFiles();
      files.addFiles(dir + File.separator + "impl" + path, "");

      Map cdl = ad.getCDLEntities();
      for (Iterator javaNames = cdl.keySet().iterator(); javaNames.hasNext();) {
        String key = (String) javaNames.next();
        String value = (String) cdl.get(key);
        String ifaceDir = (String) ifaceIndex.get(value);
        String connPath = key.replace('.', File.separatorChar);
        String connName = File.separator + connPath + ".class";
        files.addFiles(ifacePath + ifaceDir + connName, connName);
      }
      components.addFiles(info, files);
//      componentFiles.put(info, files);
//      components.put(info, info);
    }

    // look for all components stored in TR
//    enum = componentFiles.keys();
    collection = components.getInfos();
//    while (enum.hasMoreElements()) {
    for (Iterator iterator = collection.iterator(); iterator.hasNext();) { // sets all subcomponents of the initialized components
//      ComponentInfoImpl info = (ComponentInfoImpl) enum.nextElement();
      ComponentInfoImpl info = (ComponentInfoImpl) iterator.next();

      try {
        components.setSubComponents(info);
      } catch (Exception e) {
        throw new TRException("Exception thrown when initializing TR!\n" +
          "When setting subcomponents of the component:\n" + info + "\n" + e, e);
      }
    }

  }

  /**
   * Just replaces all ocurences of '::', '[' and ']' with {@link File#separator}.
   * @param clname the full name of a component.
   * @return file name of the given component.
   */
  private static String fullNameToFileName (String clname) {
    return clname.replaceAll("::|\\[|\\]", "\\" + File.separator);
/*    StringBuffer sb = new StringBuffer();
    boolean s = true;
    for (int i=0; i<clname.length(); i++) {
      if (clname.charAt(i)==':') {
        if (s) {
          sb.append(":");
          s = false;
        }
      } else {
        s = true;
        sb.append(clname.charAt(i));
      }
    }
    clname = sb.toString().replace(':', File.separatorChar);
    clname = clname.replace('[', File.separatorChar);
    clname = clname.replace(']', File.separatorChar);
    return clname;
*/
  }

  /**
   * Stores components packed in the given bundle to this TR. If a received component
   * is already stored in here, it rewrites its code with the given one. Since the
   * bundle may contain description of components only, these components are
   * (obviously) skipped.
   * @param bundle the bundle containing received components.
   * @throws TRException when some error occurs.
   */
  public synchronized void storeBundle (Bundle bundle) throws TRException {
    ComponentInfo[] compInfos = bundle.getComponents();

    System.out.println("TR: Storing a bundle containing following components:"); //todo debugging only

    for (int i = 0; i < compInfos.length; i++) {
      ComponentInfoImpl info = (ComponentInfoImpl) compInfos[i];
//      if (contains(info))
//        throw new TRException ("The component being stored to this TR is already contained in here!!\n"+info);
      ComponentFiles files = (ComponentFiles) bundle.getComponent(info);
      Hashtable filesTable = (files).getFiles();
//      System.out.println(info + "\n" + filesTable);  //tod debugging only
      String location = info.getLocation();
      if (location.startsWith("sofa://")) { // skip meta-components
        System.out.println("Description of: " + info.getName() + "[" + info.getImplementationVersion() + "]"); //todo debugging only
        continue;
      }
      String name = info.getName();
      String version = info.getImplementationVersion();
      System.out.println("Full component: " + name + "[" + version + "]"); //todo debugging only

      File file = (File) filesTable.get("/" + "index.dc");
      String indexPath = file.getPath();

      DeploymentDescriptorImpl dd = null;
      try {
        dd = new DeploymentDescriptorImpl(indexPath);
      } catch (Exception e) {
        throw new TRException("Exception thrown when storing a bundle to TR!\n" +
          "When creating the assembly descriptor of the component:\n" + info + "\n" + e, e);
      }

      Hashtable cdl = dd.getCDLEntities();

      Enumeration enum = filesTable.keys();
      while (enum.hasMoreElements()) {
        String fileName = (String) enum.nextElement(); // name of the file in a bundle
        File srcFile = (File) filesTable.get(fileName);       // file in the TEMP dir
        filesTable.remove(fileName);
        fileName = fileName.replace('/', File.separatorChar);
        if (fileName.endsWith(".class")) {
          String key = fileName.substring(1, fileName.length() - ".class".length()).replace(File.separatorChar, '.');
          if (cdl.containsKey(key)) {
            String fullName = (String) cdl.get(key); // name including specification version
            String ifaceDir;
            try {
              ifaceDir = getInterfaceDirectory(fullName);
            } catch (IOException e) {
              throw new TRException("Exception thrown when storing a bundle to TR!\n" +
                "When learning a directory where to store the interface '" + fullName + "'!\n" + e, e);
            }
            File destFile = new File(dir + File.separator + "iface" + File.separator + ifaceDir + fileName);
            ifaceIndex.put(fullName, ifaceDir);  // store the interface to the index file
            moveFile(srcFile, destFile);
            filesTable.put(fileName, destFile);
            continue; // next
          }
        }
        // it wasn't a generated file, it belongs into 'impl' directory

        File destFile = new File(dir + File.separator + "impl" + fullNameToFileName(name + "[" + version) + fileName);
        moveFile(srcFile, destFile);

        filesTable.put(fileName, destFile);
      } // for all component filesTable
      String fullName = name + "[" + version + "]";
      implIndex.put(fullName, "1");               // store the component to the index file
      info.setLocation(fullName);
      components.addFiles(info, files);
    } // for all components in the bundle

    System.out.println("End ------------------------------------------------------"); //todo debugging only

    // set subcomponents of all components in the bundle that just has been installed
    for (int i = 0; i < compInfos.length; i++) {
      ComponentInfoImpl info = (ComponentInfoImpl) compInfos[i];
      try {
        components.setSubComponents(info);
      } catch (Exception e) {
        throw new TRException("Exception thrown when storing a bundle to TR!\n" +
          "When setting subcomponents of the component:\n" + info + "\n" + e, e);
      }
    } // for all components in the bundle

    // persist the TR to file system
    try {
      saveProperties();
    } catch (IOException e) {
      throw new TRException("Exception thrown when storing a bundle to TR!\n" +
        "When saving properties!\n" + e);
    }
  }

  /**
   * Gets the directory name for an interface of a given name. Since there may be
   * more interfaces of the same name (but of different specification versions) stored
   * at TR, each of such interfaces must be stored in a different directory.
   * If an interface of the same full name (name?specification version) has already been stored
   * at this TR, it simply returns the directory. However, if the interface of the
   * same full name has not been stored at this TR, it first tries to find a directory where
   * no interface of the same (short) name is stored. If there exists such a directory, it
   * returns its name, else it creates a new directory.
   * @param fullName the full name of an interface. Full name consists of two fields delimited
   * by '?'. The first one is the hierarchical name (short name), the second one is the
   * specification version.
   * @return the directory name. It either returns name of an already created directory,
   * or it creates a new one.
   * @throws IOException when some IO operation fails.
   */
  private String getInterfaceDirectory (String fullName) throws IOException {
    synchronized (ifaceIndex) {
      String ifaceDir = (String) ifaceIndex.get(fullName);
      if (ifaceDir != null)  // is the interface stored in here?
        return ifaceDir;     // return where it is stored
      if (ifaceIndex.isEmpty()) // no interfaces are stored in here
        return getLastIndex(); // returns the last index (normally 'a').
      String name = fullName.substring(0, fullName.indexOf('?'));
      HashSet temp = new HashSet();
      temp.addAll(ifaceIndex.values()); // set of all already created directories
      Iterator iter = ifaceIndex.keySet().iterator();
      while (iter.hasNext()) { // for all names (short ones) in iface index
        String fullName2 = (String) iter.next();
        String name2 = fullName2.substring(0, fullName2.indexOf('?'));
        if (name2.equals(name)) {      // if the short name has been already stored in a directory
          String ifaceDir2 = (String) ifaceIndex.get(fullName2);
          temp.remove(ifaceDir2);      // don't use this directory again
        }
      }
      if (!temp.isEmpty())                             // has left any directory?
        return (String) temp.iterator().next();         // use it
      ifaceDir = incrementLastIndex(getLastIndex());   // otherwise create a new one
      storeLastIndex(ifaceDir);
      File d = new File(dir + File.separator + "iface" + File.separator + ifaceDir);
      d.mkdir();
      return ifaceDir;
    }
  }

  /**
   * Generates a directory name which follows after the given one. The directory names
   * are created as arabic numbers (but characters are used instead of figures). The directory name line is:
   * 'a', 'b', 'c', ... 'z', 'aa', 'ab', 'ac', ..., 'az', 'ba', ... 'zz', 'aaa', ... ,'zzz...'.
   * @param last the previous directory name.
   * @return the next directory name.
   */
  private static String incrementLastIndex (String last) {
    StringBuffer sb = new StringBuffer(last);
    int now = last.length() - 1;
    while (true) {
      if (last.charAt(now) != 'z') {
        sb.setCharAt(now, (char) (last.charAt(now) + (char) 1));
        return sb.toString();
      }
      sb.setCharAt(now, 'a');
      now--;
      if (now < 0) {
        sb.append('a');
        return sb.toString();
      }
    }
  }

  /**
   * Reads the name of an interface directory that has been used as the last one.
   * The name is stored in the file 'last' in the directory with interfaces.
   * @return the name of this directory.
   * @throws IOException when some IO operation fails.
   */
  private String getLastIndex () throws IOException {
    FileInputStream fis = new FileInputStream(dir + File.separator + "iface" + File.separator + "last");
    StringBuffer sb = new StringBuffer();
    int a = 0;
    a = fis.read();
    while (a != -1) {
      sb.append((char) a);
      a = fis.read();
    }
    fis.close();
    return sb.toString();
  }

  /**
   * Persists last index directory used for stored interfaces into a file.
   * @param last the new value of the last index.
   * @throws IOException when some IO operation fails.
   */
  private void storeLastIndex (String last) throws IOException {
    FileOutputStream fos = new FileOutputStream(dir + File.separator + "iface" + File.separator + "last");
    for (int i = 0; i < last.length(); i++) {
      fos.write((byte) last.charAt(i));
    }
    fos.close();
  }

  /**
   * Persists components and interfaces stored at this TR into 'TRdir/impl/info' and 'TRdir/iface/index' files.
   * @throws IOException when some IO operation fails.
   */
  private void saveProperties () throws IOException {
    FileOutputStream fos = new FileOutputStream(implPath + "info");
    implIndex.store(fos, null);
    fos.close();

    fos = new FileOutputStream(ifacePath + "index");
    ifaceIndex.store(fos, null);
    fos.close();
  }

  /**
   * Moves a file. First it tries to rename the file. If it fails, it simply copies
   * the content of the source file to the destination file and deletes the source file.
   * It keeps original time of last modification even in case of copying files.
   * @param srcFile the source file name (what is to be moved).
   * @param destFile the destination file (where it is to be moved).
   */
  private static void moveFile (File srcFile, File destFile) {
    destFile.getParentFile().mkdirs();
    if (srcFile.renameTo(destFile) == false) { // try use the renaming method first
      try {                                    // if failed, copy the file
        FileInputStream src = new FileInputStream(srcFile);
        FileOutputStream dest = new FileOutputStream(destFile);

        Tools.copyStream(src, dest);

        src.close();
        dest.close();
        destFile.setLastModified(srcFile.lastModified()); // keep the original time of creation
      } catch (IOException e) {
        throw new TRException("Exception thrown when moving a component file!\n" +
          "source file:" + srcFile.getName() + "\ndest file:" + destFile.getName(), e);
      }
      srcFile.delete();
    }
  }

  /**
   * Creates a bundle containing and/or describing specified components.
   * @param descs description of these components is stored into the bundle. Their code
   * is not.
   * @param comps description as well as code of these components is stored into the bundle.
   * @param inferiors if <code>true</code> all inferiors (direct as well as indirect) of the
   * specified components are included in the returned bundle.
   * @return bundle containing specified components or their description.
   * @throws TRException when a required component is not found on this TR.
   */
  public synchronized Bundle getBundle (ComponentInfo[] descs, ComponentInfo[] comps, boolean inferiors) throws TRException {
    if (inferiors) {
      descs = getTransitiveClosure((ComponentInfoImpl[]) descs);
      comps = getTransitiveClosure((ComponentInfoImpl[]) comps);
    }

    System.out.println("TR: Getting a bundle containing following components:"); //todo debugging only

    BundleImpl bundle = new BundleImpl();
    if (comps != null) {
      for (int i = 0; i < comps.length; i++) {
        ComponentInfo info = comps[i];
        System.out.println("Full component: " + info.getName() + "[" + info.getImplementationVersion() + "]"); //todo debugging only
        ComponentFiles files = components.getFiles(info);
        if (files == null) throw new TRException("Required component not stored in this TR!" + info);
        bundle.add((ComponentInfoImpl) info, files);
      }
    }
    if (descs != null) {
      for (int i = 0; i < descs.length; i++) {
        ComponentInfo info = descs[i];
        System.out.println("Description of: " + info.getName() + "[" + info.getImplementationVersion() + "]"); //todo debugging only
        bundle.add((ComponentInfoImpl) info, BundleImpl.NO_FILES);
      }
    }
    System.out.println("End ------------------------------------------------------"); //todo debugging only
    return bundle;
  }

  private ComponentInfoImpl[] getTransitiveClosure (ComponentInfoImpl[] comps) {
    if (comps == null)
      return null;
    // list for iterating through all components and their (in the current time) known subcomponents
    List list = new LinkedList();
    // set for learning if a concrete component is in the list
    Set set = new HashSet();

    for (int i = 0; i < comps.length; i++) {
      ComponentInfoImpl comp = components.getInfo(comps[i]);  // use the instance filled not only with "key properties"
      list.add(comp);
      set.add(comp);
    }

    for (int i = 0; i < list.size(); i++) {
      ComponentInfoImpl info = (ComponentInfoImpl) list.get(i);
      ComponentInfo[] subs = info.getSubComponents();
      if (subs != null) {
        for (int j = 0; j < subs.length; j++) {
          ComponentInfo sub = subs[j];
          if (!set.contains(sub)) {  // the component is not in the set and thus in the list either
            list.add(sub);
            set.add(sub);
          }
        }
      }
    }

    return (ComponentInfoImpl[]) list.toArray(new ComponentInfoImpl[list.size()]);
  }

  /**
   * Returns <code>true</code> if the specified component is stored in this TR.
   * Only two fields in <code>ComponentInfo</code> must be filled -
   * the {@link ComponentInfo#getName component name} and it's
   * {@link ComponentInfo#getImplementationVersion implementation version}.
   * @param desc the component which is being asked.
   * @return <code>true</code> if the component is in TR, <code>false</code> otherwise.
   */
  public synchronized boolean contains (ComponentInfo desc) {
    return components.contains(desc);
  }

  /**
   * Lists all components stored on this TR.
   * @return a bundle containing all components stored on this TR.
   */
  public Bundle list () {
    Set infos = components.getInfos();
    ComponentInfo[] descs = (ComponentInfo[]) infos.toArray(new ComponentInfo[infos.size()]);
    return getBundle(descs, null, false);
  }

}
