/* $Id: Resolver.java,v 1.1 2003/03/17 10:56:19 hnetynka Exp $ */
package SOFA.Tools.Resolver;

import org.objectweb.asm.*;

/** SOFA class name resolver. Tool for augmentation of Java class files.
  * 
  * @author Petr Hnetynka
  */
public class Resolver {
  
  private java.util.Hashtable trTable;
  
  /** Constucts new Resolver.
    *
    * @param table   translation table with old names as keys
    */
  public Resolver(java.util.Hashtable table) {
    trTable = table;
  }

  /** Process byte code augmention
    * 
    * @param is   input stream with byte code
    * @return     augmented byte code
    * @throws IOException IO exception
    */
  public byte[] process(java.io.InputStream is) throws java.io.IOException {
    ClassReader cr = new ClassReader(is);
    ClassWriter cw = new ClassWriter(false);
    ClassVisitor cv = new SOFAClassAdapter(cw, trTable);
    cr.accept(cv, false);
    return cw.toByteArray();
  }

  /** Converts class name to class name used in the Java VM. Indeed, it replaces dots by slashes.
    * E.g.   java.lang.String  -&gt; java/lang/String
    * 
    * @param className    class name
    * @return             converted class name
    */
  public static String classNameToVMClassName(String className) {
    StringBuffer ret = new StringBuffer();
    for (int i=0; i<className.length(); i++) {
      if (className.charAt(i) == '.')
        ret.append('/');
      else
        ret.append(className.charAt(i));
    }
    return ret.toString();
  }

  /** Converts class name of cdl definition to class name used in the Java VM.
    * E.g.   SOFA.demos.logdemo.LogInterface + ver: nenya.ms.mff.cuni.cz!0  -&gt;
    *        SOFA/demos/logdemo/LogInterface_Q_nenya_ms_mff_cuni_cz_E_0
    * 
    * @param cdlClassName   class name
    * @param version        version of cdl definition
    * @return               converted classname
    */
  public static String cdlClassNameToVMClassName(String cdlClassName, String version) {
    StringBuffer ret = new StringBuffer();
    for (int i=0; i<cdlClassName.length(); i++) {
      if (cdlClassName.charAt(i) == '.')
        ret.append('/');
      else
        ret.append(cdlClassName.charAt(i));
    }
    ret.append("_Q_");
    for (int i=0; i<version.length(); i++) {
      switch (version.charAt(i)) {
      case '.': 
        ret.append('_');
        break;
      case '!':
        ret.append("_E_");
        break;
      default:
        ret.append(version.charAt(i));
      }
    }
    return ret.toString();
  }
}

/** Performs augmentation of classes.
  * 
  * @author Petr Hnetynka
  */
class SOFAClassAdapter extends ClassAdapter implements Constants {
  
  private java.util.Hashtable trTable;
  
  /** Constructor
    */
  public SOFAClassAdapter (ClassVisitor cv, java.util.Hashtable table) {
    super(cv);
    trTable = table;
  }

  public void visit ( final int access, final String name, final String superName, final String[] interfaces, final String sourceFile) {
    String newName = (String) trTable.get(name);
    if (newName == null) {
 //     System.out.println("Warning: class \""+name+"\" was not augmented.");
      newName = name;
    }
    String newSuperName = (String) trTable.get(superName);
    newSuperName = (newSuperName != null) ? newSuperName : superName;
    String [] newInterfaces = new String [interfaces.length];
    for (int i=0; i<interfaces.length; i++) {
      newInterfaces[i] = (String) trTable.get(interfaces[i]);
      newInterfaces[i] = (newInterfaces[i] != null) ? newInterfaces[i] : interfaces[i];
    }
    cv.visit(access, newName, newSuperName, newInterfaces, sourceFile);
  }

  public void visitField(final int access, final String name, final String desc, final Object value) {
    String newDesc = augmentSimpleDesc(desc, trTable);
    cv.visitField(access, name, newDesc, value);
  }

  public CodeVisitor visitMethod(final int access, final String name, final String desc, final String[] exceptions) {
    String [] newExceptions = null;
    if (exceptions != null) { // augment exceptions
      newExceptions = new String [exceptions.length];
      for (int i=0; i<exceptions.length; i++) {
        newExceptions[i] = (String) trTable.get(exceptions[i]);
        newExceptions[i] = (newExceptions[i] != null) ? newExceptions[i] : exceptions[i];
      }
    }
    // augment description
    // description has format   (parameteres_description)ret_type_description
    String paramTypes = desc.substring(1,desc.indexOf(')'));
    String retType = desc.substring(desc.indexOf(')')+1);
    String newRetType = augmentSimpleDesc(retType, trTable);
    String newParamTypes = augmentListOfDescs(paramTypes, trTable);
    String newDesc = "(" + newParamTypes + ")" + newRetType;
    CodeVisitor mv = cv.visitMethod(access, name, newDesc, newExceptions);
    return mv == null ? null : new TraceFieldCodeAdapter(mv, trTable);
  }

  /** Performs augmentation of simple description of type.
    * 
    * @param desc     original type description
    * @param trTable  translation table
    * @return         augmented type description
    */
  public static String augmentSimpleDesc(String desc, java.util.Hashtable trTable) {
    StringBuffer type = new StringBuffer();
    StringBuffer prefix = new StringBuffer();
    int i=0;
    while (desc.charAt(i) == '[') {  // skip field dcl
      prefix.append('[');
      i++;
    }
    if (desc.charAt(i) == 'L') {
      i++; // skip 'L'
      while (desc.charAt(i) != ';') { // read type
        type.append(desc.charAt(i++));
      }
      String newType = (String) trTable.get(type.toString());
      newType = (newType != null) ? newType : type.toString();
      return prefix.toString() + "L" + newType + ";";
    } else {
      return desc;
    }
  }

  /** Perform augmentation of list of descs (from method decl.).
    * 
    * @param desc     list of descs
    * @param trTable  translation table
    * @return         augmented list of descs
    */
  public static String augmentListOfDescs(String descs, java.util.Hashtable trTable) {
    StringBuffer ret = new StringBuffer();
    StringBuffer cur = new StringBuffer();
    boolean typeNow = false;
    for (int i=0; i<descs.length(); i++) {
      if (typeNow) {
        if (descs.charAt(i) != ';') {
          cur.append(descs.charAt(i));
        } else {
          String newType = (String) trTable.get(cur.toString());
          newType = (newType != null) ? newType : cur.toString();
          ret.append(newType);
          ret.append(';');
          cur = new StringBuffer();
          typeNow = false;
        }
      } else {
        if (descs.charAt(i) == 'L') {
          typeNow = true;
        }
        ret.append(descs.charAt(i));
      }
    }
    return ret.toString();
  }
}

/** Goes through method's instructions and performs augmentation of references.
  * 
  * @author Petr Hnetynka
  */
class TraceFieldCodeAdapter extends CodeAdapter implements Constants {
  private java.util.Hashtable trTable;
  public TraceFieldCodeAdapter (CodeVisitor cv, java.util.Hashtable table) {
    super(cv);
    trTable = table;
  }

  public void visitFieldInsn(final int opcode, final String owner, final String name, final String desc) {
    String newOwner = (String) trTable.get(owner);
    newOwner = (newOwner != null) ? newOwner : owner; 
    String newDesc = SOFAClassAdapter.augmentSimpleDesc(desc, trTable);
    cv.visitFieldInsn(opcode, newOwner, name, newDesc);
  }

  public void visitMethodInsn(int opcode, String owner, String name, String desc) {
    String paramTypes = desc.substring(1,desc.indexOf(')'));
    String retType = desc.substring(desc.indexOf(')')+1);
    String newRetType = SOFAClassAdapter.augmentSimpleDesc(retType, trTable);
    String newParamTypes = SOFAClassAdapter.augmentListOfDescs(paramTypes, trTable);
    String newDesc = "(" + newParamTypes + ")" + newRetType;
    String newOwner = (String) trTable.get(owner);
    newOwner = (newOwner != null) ? newOwner : owner;
    cv.visitMethodInsn(opcode, newOwner, name, newDesc);
  }

  public void visitTypeInsn(int opcode, String desc) {
    String newDesc = (String) trTable.get(desc);
    newDesc = (newDesc != null) ? newDesc : desc;
    cv.visitTypeInsn(opcode, newDesc);
  }

  public void visitLocalVariable(String name, String desc, Label start, Label end, int index) {
    String newDesc = SOFAClassAdapter.augmentSimpleDesc(desc, trTable);
    cv.visitLocalVariable(name, newDesc, start, end, index);
  }

  public void visitMultiANewArrayInsn(String desc, int dims) {
    String newDesc = SOFAClassAdapter.augmentSimpleDesc(desc, trTable);
    cv.visitMultiANewArrayInsn(newDesc, dims);
  }
  
  public void visitTryCatchBlock(Label start, Label end, Label handler, String type) {
    String newType = null;
    if (type != null) {
      newType = (String) trTable.get(type);
      newType = (newType != null) ? newType : type; 
    }
    cv.visitTryCatchBlock(start, end, handler, newType);
  }
}
